/* Copyright (c) 2006 MetaCarta, Inc., published under a modified BSD license.
 * See http://svn.openlayers.org/trunk/openlayers/repository-license.txt 
 * for the full text of the license. */


/**
 * @class
 * @requires OpenLayers/Map.js
 */
OpenLayers.Layer = OpenLayers.Class.create();
OpenLayers.Layer.prototype = {

    /** @type String */
    id: null,

    /** @type String */
    name: null,

    /** @type DOMElement */
    div: null,

    /** supported application event types
     * 
     * @type Array */
    EVENT_TYPES: [ 
        "loadstart", "loadend", "loadcancel"],

    /** @type OpenLayers.Events */
    events: null,

    /** This variable is set when the layer is added to the map, via the 
     *  accessor function setMap()
     * 
     * @type OpenLayers.Map */
    map: null,
    
    /** Whether or not the layer is a base layer. This should be set 
     *   individually by all subclasses. 
     *   Default is false
     * 
     * @type Boolean
     */
    isBaseLayer: false,
 
    /** asserts whether or not the layer's images have an alpha channel 
     * 
     * @type boolean */
    alpha: false,

    /** should the layer's name appear in the layer switcher?
     * 
     * @type boolean */
    displayInLayerSwitcher: true,

    /** Whether or not the layer should be displayed in the map
     * 
     * @type Boolean
     */
    visibility: true,

    /** Whether or not the map's current resolution is within this layer's
     *   min/max range -- this is set in map's setCenter() whenever zoom
     *   changes
     * 
     * @type Boolean
     */
    inRange: false,
    
    /**
     * For layers that use Tile.Image, the image size is cached here.  For
     * layers without a gutter, the image size is equal to the tile size.
     * For layers with a gutter, the image is larger than the tile by twice
     * the gutter in each dimension.
     * 
     * @type OpenLayers.Size
     * @private
     */
    imageSize: null,
    
    /**
     * For layers that use Tile.Image, the image offset is cached here.
     * Layers without a gutter have zero offset.  For layers with a gutter,
     * the image offset represents displacement due to the gutter.
     * 
     * @type OpenLayers.Pixel
     * @private
     */
    imageOffset: null,

  // OPTIONS

    /** @type Array */
    options: null,

    /** Determines the width (in pixels) of the gutter around image tiles 
     * to ignore.  By setting this property to a non-zero value, images 
     * will be requested that are wider and taller than the tile size by 
     * a value of 2 x gutter.  This allows artifacts of rendering at tile 
     * edges to be ignored.  Set a gutter value that is equal to half the size 
     * of the widest symbol that needs to be displayed.  Defaults to zero.
     * Non-tiled layers always have zero gutter.
     *  
     * @type Int 
    */ 
    gutter: 0, 

    /** @type String */
    projection: null,    
    
    /** @type String */
    units: null,

    /** @type Array */
    scales: null,

    /** @type Array */
    resolutions: null,
    
    /** @type OpenLayers.Bounds */
    maxExtent: null,
    
    /** @type OpenLayers.Bounds */
    minExtent: null,
    
    /** @type float */
    maxResolution: null,

    /** @type float */
    minResolution: null,

    /** @type int */
    numZoomLevels: null,
   
    /** @type float */
    minScale: null,
    
    /** @type float */
    maxScale: null,

    /** @type Boolean */
    displayOutsideMaxExtent: false,
    
    /** did the image finish loading before a new draw was initiated?
     * @type Boolean */
    doneLoading: false,

    /** @type Boolen */
    firstCall: true, 


    /**
     * @constructor
     * 
     * @param {String} name
     * @param {Object} options Hashtable of extra options to tag onto the layer
     */
    initialize: function(name, options) {

        this.addOptions(options);

        this.name = name;
        
        if (this.id == null) {

            this.id = OpenLayers.Util.createUniqueID(this.CLASS_NAME + "_");

            this.div = OpenLayers.Util.createDiv();
            this.div.style.width = "100%";
            this.div.style.height = "100%";
            this.div.id = this.id;

            this.events = new OpenLayers.Events(this, this.div, 
                                                this.EVENT_TYPES);
            this.events.register("loadend", this, this.setLoadendVisibility);
        }
        
    },
    
    /**
     * Destroy is a destructor: this is to alleviate cyclic references which
     * the Javascript garbage cleaner can not take care of on its own.
     * 
     * @param {Boolean} setNewBaseLayer Should a new baselayer be selected when
     *                                  this has been removed?
     *                                  Default is true
     */
    destroy: function(setNewBaseLayer) {
        if (setNewBaseLayer == null) {
            setNewBaseLayer = true;
        }
        if (this.map != null) {
            this.map.removeLayer(this, setNewBaseLayer);
        }
        this.map = null;
        this.name = null;
        this.div = null;
        this.options = null;

        if (this.events) {
            this.events.destroy();
        }
        this.events = null;
    },
    
   /**
    * @returns An exact clone of this OpenLayers.Layer
    * @type OpenLayers.Layer
    */
    clone: function (obj) {
        
        if (obj == null) {
            obj = new OpenLayers.Layer(this.name, this.options);
        } 
        
        // catch any randomly tagged-on properties
        OpenLayers.Util.applyDefaults(obj, this);
        
        // a cloned layer should never have its map property set
        //  because it has not been added to a map yet. 
        obj.map = null;
        
        return obj;
    },
    
    /** 
     * @param {String} newName
     */
    setName: function(newName) {
        if (newName != this.name) {
            this.name = newName;
            if (this.map != null) {
                this.map.events.triggerEvent("changelayer");
            }
        }
    },    
    
   /**
    * @param {Object} newOptions
    */
    addOptions: function (newOptions) {
        
        if (this.options == null) {
            this.options = new Object();
        }
        
        // update our copy for clone
        OpenLayers.Util.extend(this.options, newOptions);

        // add new options to this
        OpenLayers.Util.extend(this, newOptions);
    },
    
    /**
     * 
     */
    onMapResize: function() {
        //this function can be implemented by subclasses  
    },

    /**
     * @param {OpenLayers.Bounds} bound
     * @param {Boolean} zoomChanged tells when zoom has changed, as layers 
     *                   have to do some init work in that case.
     * @param {Boolean} dragging
     */
    moveTo:function(bounds, zoomChanged, dragging) {
        var display = this.visibility;
        if (!this.isBaseLayer) {
            display = display && this.inRange;
        }
        this.display(display);
    },

    /** Set the map property for the layer. This is done through an accessor
     *   so that subclasses can override this and take special action once 
     *   they have their map variable set. 
     * 
     *  Here we take care to bring over any of the necessary default properties
     *   from the map. 
     * 
     * @param {OpenLayers.Map} map
     */
    setMap: function(map) {
        if (this.map == null) {
        
            this.map = map;
            
            // grab some essential layer data from the map if it hasn't already
            //  been set
            this.maxExtent = this.maxExtent || this.map.maxExtent;
            this.projection = this.projection || this.map.projection;
            this.units = this.units || this.map.units;
            
            this.initResolutions();
            
            if (!this.isBaseLayer) {
                this.inRange = this.calculateInRange();
                var show = ((this.visibility) && (this.inRange));
                this.div.style.display = show ? "" : "none";
            }
            
            // deal with gutters
            this.setTileSize();
        }
    },
  
    /**
     * Set the tile size based on the map size.  This also sets layer.imageSize
     * and layer.imageOffset for use by Tile.Image.
     *
     * @param OpenLayers.Size
     */
    setTileSize: function(size) {
        var tileSize = (size) ? size :
                                ((this.tileSize) ? this.tileSize :
                                                   this.map.getTileSize());
        this.tileSize = tileSize;
        if(this.gutter) {
            // layers with gutters need non-null tile sizes
            //if(tileSize == null) {
            //    OpenLayers.console.error("Error in layer.setMap() for " +
            //                              this.name + ": layers with gutters " +
            //                              "need non-null tile sizes");
            //}
            this.imageOffset = new OpenLayers.Pixel(-this.gutter, -this.gutter); 
            this.imageSize = new OpenLayers.Size(tileSize.w + (2 * this.gutter), 
                                                 tileSize.h + (2 * this.gutter)); 
        } else {
            // layers without gutters may have null tile size - as long
            // as they don't rely on Tile.Image
            this.imageSize = tileSize;
            this.imageOffset = new OpenLayers.Pixel(0, 0);
        }
    },

    /**
    * @returns Whether or not the layer should be displayed (if in range)
    * @type Boolean
    */
    getVisibility: function() {
        return this.visibility;
    },

    /** Set the visibility flag for the layer and hide/show&redraw accordingly. 
     *   Fire event unless otherwise specified
     * 
     * Note that visibility is no longer simply whether or not the layer's
     *  style.display is set to "block". Now we store a 'visibility' state 
     *  property on the layer class, this allows us to remember whether or not
     *  we *desire* for a layer to be visible. In the case where the map's 
     *  resolution is out of the layer's range, this desire may be subverted.
     *  
     * @param {Boolean} visible Whether or not to display the layer 
     *                          (if in range)
     * @param {Boolean} noEvent
     */
    setVisibility: function(visibility, noEvent) {
        if (visibility != this.visibility) {
            this.visibility = visibility;
            this.display(visibility);
            if (visibility && this.map != null) {
                var extent = this.map.getExtent();
                if (extent != null) {
                    this.moveTo(extent, true);
                }
            }
            if ((this.map != null) && 
                ((noEvent == null) || (noEvent == false))) {
                this.map.events.triggerEvent("changelayer");
            }
        }
    },

    /** Hide or show the Layer
     * 
     * @param {Boolean} display
     */
    display: function(display) {
        if (display != (this.div.style.display != "none")) {
            this.div.style.display = (display) ? "block" : "none";
        }
    },

    /**
     * @returns Whether or not the layer is displayable at the current map's
     *          current resolution
     * @type Boolean
     */
    calculateInRange: function() {
        var inRange = false;
        if (this.map) {
            var resolution = this.map.getResolution();
            inRange = ( (resolution >= this.minResolution) &&
                        (resolution <= this.maxResolution) );
        }
        return inRange;
    },

    /** 
     * @param {Boolean} isBaseLayer 
     */
    setIsBaseLayer: function(isBaseLayer) {
        if (isBaseLayer != this.isBaseLayer) {
            this.isBaseLayer = isBaseLayer;
            if (this.map != null) {
                this.map.events.triggerEvent("changelayer");
            }
        }
    },

  /********************************************************/
  /*                                                      */
  /*                 Baselayer Functions                  */
  /*                                                      */
  /********************************************************/
  
    /** This method's responsibility is to set up the 'resolutions' array 
     *   for the layer -- this array is what the layer will use to interface
     *   between the zoom levels of the map and the resolution display of the
     *   layer.
     * 
     *  The user has several options that determine how the array is set up.
     *  
     *  For a detailed explanation, see the following wiki from the 
     *   openlayers.org homepage:
     * 
     *  http://trac.openlayers.org/wiki/SettingZoomLevels
     * 
     * @private
     */
    initResolutions: function() {

        // These are the relevant options which are used for calculating 
        //  resolutions information.
        //
        var props = new Array(
          'projection', 'units',
          'scales', 'resolutions',
          'maxScale', 'minScale', 
          'maxResolution', 'minResolution', 
          'minExtent', 'maxExtent',
          'numZoomLevels', 'maxZoomLevel'
        );

        // First we create a new object where we will store all of the 
        //  resolution-related properties that we find in either the layer's
        //  'options' array or from the map.
        //
        var confProps = new Object();        
        for(var i=0; i < props.length; i++) {
            var property = props[i];
            confProps[property] = this.options[property] || this.map[property];
        }

        // If numZoomLevels hasn't been set and the maxZoomLevel *has*, 
        //  then use maxZoomLevel to calculate numZoomLevels
        //
        if ( (!confProps.numZoomLevels) && (confProps.maxZoomLevel) ) {
            confProps.numZoomLevels = confProps.maxZoomLevel + 1;
        }

        // First off, we take whatever hodge-podge of values we have and 
        //  calculate/distill them down into a resolutions[] array
        //
        if ((confProps.scales != null) || (confProps.resolutions != null)) {
          //preset levels
            if (confProps.scales != null) {
                confProps.resolutions = new Array();
                for(var i = 0; i < confProps.scales.length; i++) {
                    var scale = confProps.scales[i];
                    confProps.resolutions[i] = 
                       OpenLayers.Util.getResolutionFromScale(scale, 
                                                              confProps.units);
                }
            }
            confProps.numZoomLevels = confProps.resolutions.length;

        } else {
          //maxResolution and numZoomLevels based calculation
            
            confProps.resolutions = new Array();
            
            // determine maxResolution
            if (confProps.minScale) {
                confProps.maxResolution = 
                    OpenLayers.Util.getResolutionFromScale(confProps.minScale, 
                                                           confProps.units);
            } else if (confProps.maxResolution == "auto") {
                var viewSize = this.map.getSize();
                var wRes = confProps.maxExtent.getWidth() / viewSize.w;
                var hRes = confProps.maxExtent.getHeight()/ viewSize.h;
                confProps.maxResolution = Math.max(wRes, hRes);
            } 

            // determine minResolution
            if (confProps.maxScale != null) {           
                confProps.minResolution = 
                    OpenLayers.Util.getResolutionFromScale(confProps.maxScale);
            } else if ( (confProps.minResolution == "auto") && 
                        (confProps.minExtent != null) ) {
                var viewSize = this.map.getSize();
                var wRes = confProps.minExtent.getWidth() / viewSize.w;
                var hRes = confProps.minExtent.getHeight()/ viewSize.h;
                confProps.minResolution = Math.max(wRes, hRes);
            } 

            // determine numZoomLevels
            if (confProps.minResolution != null) {
                var ratio = confProps.maxResolution / confProps.minResolution;
                confProps.numZoomLevels = 
                    Math.floor(Math.log(ratio) / Math.log(2)) + 1;
            }
            
            // now we have numZoomLevels and maxResolution, 
            //  we can populate the resolutions array
            for (var i=0; i < confProps.numZoomLevels; i++) {
                var res = confProps.maxResolution / Math.pow(2, i)
                confProps.resolutions.push(res);
            }    
        }
        
        //sort resolutions array ascendingly
        //
        confProps.resolutions.sort( function(a, b) { return(b-a); } );

        // now set our newly calculated values back to the layer 
        //  Note: We specifically do *not* set them to layer.options, which we 
        //        will preserve as it was when we added this layer to the map. 
        //        this way cloned layers reset themselves to new map div 
        //        dimensions)
        //

        this.resolutions = confProps.resolutions;
        this.maxResolution = confProps.resolutions[0];
        var lastIndex = confProps.resolutions.length - 1;
        this.minResolution = confProps.resolutions[lastIndex];
        
        this.scales = new Array();
        for(var i = 0; i < confProps.resolutions.length; i++) {
            this.scales[i] = 
               OpenLayers.Util.getScaleFromResolution(confProps.resolutions[i], 
                                                      confProps.units);
        }
        this.minScale = this.scales[0];
        this.maxScale = this.scales[this.scales.length - 1];
        
        this.numZoomLevels = confProps.numZoomLevels;
    },

    /**
     * @returns The currently selected resolution of the map, taken from the
     *          resolutions array, indexed by current zoom level.
     * @type float
     */
    getResolution: function() {
        var zoom = this.map.getZoom();
        if (zoom >= this.map.getNumZoomLevels())
            zoom = this.map.getNumZoomLevels() - 1;
        return this.resolutions[zoom];
    },

    /** 
     * @returns A Bounds object which represents the lon/lat bounds of the 
     *          current viewPort.
     * @type OpenLayers.Bounds
     */
    getExtent: function() {
        // just use stock map calculateBounds function -- passing no arguments
        //  means it will user map's current center & resolution
        //
        return this.map.calculateBounds();
    },

    /**
     * @param {OpenLayers.Bounds} bounds
     *
     * @returns The index of the zoomLevel (entry in the resolutions array) 
     *           that still contains the passed-in extent. We do this by 
     *           calculating the ideal resolution for the given exteng (based
     *           on the map size) and then find the smallest resolution that 
     *           is greater than this ideal resolution.
     * @type int
     */
    getZoomForExtent: function(extent) {
        var viewSize = this.map.getSize();
        var idealResolution = Math.max( extent.getWidth()  / viewSize.w,
                                        extent.getHeight() / viewSize.h );

        return this.getZoomForResolution(idealResolution);
    },
    
    /**
     * @param {float} resolution
     *
     * @returns The index of the zoomLevel (entry in the resolutions array) 
     *           that is the smallest resolution that is greater than the 
     *           passed-in resolution.
     * @type int
     */
    getZoomForResolution: function(resolution) {
        
        for(var i=1; i < this.resolutions.length; i++) {
            if ( this.resolutions[i] < resolution) {
                break;
            }
        }
        return (i - 1);
    },
    
    /**
     * @param {OpenLayers.Pixel} viewPortPx
     *
     * @returns An OpenLayers.LonLat which is the passed-in view port
     *          OpenLayers.Pixel, translated into lon/lat by the layer
     * @type OpenLayers.LonLat
     */
    getLonLatFromViewPortPx: function (viewPortPx) {
        var lonlat = null;
        if (viewPortPx != null) {
            var size = this.map.getSize();
            var center = this.map.getCenter();
            if (center) {
                var res  = this.map.getResolution();

                var delta_x = viewPortPx.x - Math.ceil(size.w / 2);
                var delta_y = viewPortPx.y - Math.ceil(size.h / 2);

            
                lonlat = new OpenLayers.LonLat(center.lon + delta_x * res ,
                                             center.lat - delta_y * res); 
            } // else { DEBUG STATEMENT }
        }
        return lonlat;
    },

    /**
     * @param {OpenLayers.LonLat} lonlat
     *
     * @returns An OpenLayers.Pixel which is the passed-in OpenLayers.LonLat, 
     *          translated into view port pixels
     * @type OpenLayers.Pixel
     */
    getViewPortPxFromLonLat: function (lonlat, resolution) {
        var px = null; 
        if (lonlat != null) {
            if (!resolution)
                resolution = this.map.getResolution();
            var extent = this.map.getExtent();
            px = new OpenLayers.Pixel(
                           Math.round(1/resolution * (lonlat.lon - extent.left)),
                           Math.round(1/resolution * (extent.top - lonlat.lat))
                           );    
        }
        return px;
    },
    
    /**
     * Adjusts the extent of a bounds in map units by the layer's gutter
     * in pixels.
     * 
     * @param {OpenLayers.Bounds} bounds
     * @type OpenLayers.Bounds
     * @return A bounds adjusted in height and width by the gutter
     */
    adjustBoundsByGutter: function(bounds) {
        var mapGutter = this.gutter * this.map.getResolution();
        bounds = new OpenLayers.Bounds(bounds.left - mapGutter,
                                       bounds.bottom - mapGutter,
                                       bounds.right + mapGutter,
                                       bounds.top + mapGutter);
        return bounds;
    },
    
    /**
     * Sets the opacity for the entire layer (all images)
     * @param {Float} opacity
     */
    setOpacity: function(opacity) {
        if (opacity != this.opacity) {
            this.opacity = opacity;
            for(var i=0; i<this.div.childNodes.length; ++i) {
                var element = this.div.childNodes[i];
                OpenLayers.Util.modifyDOMElement(element, null, null, null, 
                                                 null, null, null, opacity);
            }
        }
    },

    /**
    * @param {int} zIdx
    * @private
    */    
    setZIndex: function (zIdx) {
        this.div.style.zIndex = zIdx;
    },




    /********************************************************/
    /*                                                      */
    /*       Baselayer Functions for zooming/scaling        */
    /*                                                      */
    /********************************************************/

    /** 
     * Initializes zoomOut tile and sets a x-times bigger tile size.
     *
     * Default function; will override in subclasses,
     * otherwise it returns false
     * 
     * @returns false if no zoomOutTile is set
     * @type Boolean
     */
    setZoomOutTile:function() {
        return false;
    },

    /**
     * Initializes zoomOut tile and sets a x-times bigger tile size.
     * 
     * special function; shares with some baselayers; calls from subclasses
     *
     */
    setZoomOutTile_share:function() {
        if (!this.zoomOutTile) {
            var bounds = this.map.getMaxExtent();
            var resolution = this.resolutions[0]; 

            // calculate tile bounds
            var n = 0;
            do {
                resolution = resolution * Math.pow(2, n);
                
                var tilelon = resolution * this.map.tileSize.w;
                var tilelat = resolution * this.map.tileSize.h;
            
                var tileoffsetlon = bounds.left;
                var tileoffsetlat = bounds.bottom;
                
                var tileBounds = new OpenLayers.Bounds(tileoffsetlon, 
                                                       tileoffsetlat, 
                                                       tileoffsetlon + tilelon,
                                                       tileoffsetlat + tilelat);
                n++;
            }
            while (!tileBounds.containsBounds(bounds, false, true));

            var tileSize = new OpenLayers.Size();
            tileSize.w = this.map.zoomOutTileSizeFactor * this.map.tileSize.w;
            tileSize.h = this.map.zoomOutTileSizeFactor * this.map.tileSize.h;

            // formulate request url string
            var url = this.getURL(tileBounds, tileSize); 
            // default start position
            var pos = new OpenLayers.Pixel (0,0); 
            // default start size
            var tilesize = new OpenLayers.Size(0,0);

            this.zoomOutTile = new OpenLayers.Tile.Image(this, pos, tileBounds, 
                                                        url, tileSize);

            // draw tile -> sets imgDiv
            this.zoomOutTile.draw();

            // resize zoomOutTile
            this.zoomOutTile.imgDiv.style.width = 0;
            this.zoomOutTile.imgDiv.style.height = 0;
            
            // set zoomOut tile invisible 
            if (this.map.baseLayer.zoomOutTile)
                this.map.baseLayer.zoomOutTile.imgDiv.style.display = "none";
        }
    },
 
    /** 
     * Gets the current tile size of the layer.
     *
     * Default function; will override in subclasses,
     * otherwise it returns false - that means: no scaling possible!
     * 
     * @returns null if no tile size set
     * @type OpenLayers.Size
     */
    getTileSize:function() {
        return null;
    },


    /** 
     * Gets tile, which contains the center of the viewport.
     *
     * Default function; will override in subclasses,
     * otherwise it returns null - that means: no scaling possible!
     * 
     * @returns returns null if no center tile set
     * @type OpenLayers.Tile
     */
    getCenterTile:function() {
        return null;
    },


    /** 
     * Sets all tiles outside the viewport invisible. 
     * Only for baselayers with a grid of tiles.
     *
     * Default function; will override in subclasses,
     * otherwise it returns false
     * 
     * @returns false if baselayer has no grid
     * @type Boolean
     */
    setTilesOutsideInvisible:function() {
        return false;
    },

    /**
     * Scales tile to the new tile size (changes size and position).
     * If you have a grid, the tile is the center tile. Otherwise it's
     * the only one.
     * 
     * @param {OpenLayers.Tile} tile
     * @param {OpenLayers.Size} newTileSize
     * 
     * @returns true if tile is scaled, otherwise false
     * @type Boolean
     */
    scaleTileTo:function(tile, newTileSize) {
        //convert the center of the map in pixel
        var centerPx = this.map.getPixelFromLonLat(this.map.getCenter());

        //reposition and resize the tile 
        if (tile) {
            if ((newTileSize.w < Math.pow(2,15)) &&
                (newTileSize.h < Math.pow(2,15)) ){

                // set real start position of centerTile
                var startPosX = (tile.position.x +
                  parseInt(this.map.layerContainerDiv.style.left));
                // set new left position
                var leftOffset = newTileSize.w / tile.size.w
                  * (centerPx.x - startPosX);
                leftOffset = Math.round(leftOffset);
                var newLeftPos = centerPx.x - leftOffset
                  - parseInt(this.map.layerContainerDiv.style.left);
                tile.imgDiv.style.left = Math.round(newLeftPos) + "px";

                // set real start position of centerTile
                var startPosY = (tile.position.y +
                  parseInt(this.map.layerContainerDiv.style.top));
                //set new top position
                var topOffset = newTileSize.h / tile.size.h
                  * (centerPx.y - startPosY);
                topOffset = Math.round(topOffset);
                var newTopPos = centerPx.y - topOffset 
                  - parseInt(this.map.layerContainerDiv.style.top);
                tile.imgDiv.style.top = Math.round(newTopPos) + "px";

                //set new width and height
                tile.imgDiv.style.width = Math.round(newTileSize.w) + "px";
                tile.imgDiv.style.height = Math.round(newTileSize.h) + "px";
            }
            return true;
        }
        else {
            return false;
        }
    }, 

    /** 
     * Scales all tiles of a grid to the new tile size 
     * (changes size and position).
     *
     * Default function; will override in subclasses,
     * otherwise it returns false
     * 
     * @param {OpenLayers.Tile} centerTile
     * @param {OpenLayers.Size} newTileSize
     *
     * @returns false if no grid is scaled
     * @type Boolean
     */
    scaleTilesOfGrid:function(centertile, newTileSize) {
        return false;
    },


    /**
     * Reposition and resize the zoomOut tile to prevent a white frame
     * during zoomOut.
     *
     * @param {float} newZoomlevel
     * @param {float} newResolution
     *
     * @returns false if no zoomOutTile scaled
     * @type Boolean
     */
    scaleZoomOutTile: function(newZoomlevel, newResolution) {
        return false;
    },

    /**
     * The actual zoomOutTile scale function:
     * Reposition and resize the zoomOut tile to prevent a white frame
     * during zoomOut.
     *
     * special function; shares with some baselayers; calls from subclasses
     *
     * @param {float} newZoomlevel
     * @param {float} newResolution
     *
     * @returns true after the zoomOutTile is scaled
     * @type Boolean
     */
    scaleZoomOutTile_share: function(newZoomlevel, newResolution) {

        //convert the current center of the map in pixel
        var centerPx = this.map.getPixelFromLonLat(this.map.getCenter());


        var newTileSize = new OpenLayers.Size();
        newTileSize.w = Math.round(this.map.tileSize.w * Math.pow(2, newZoomlevel));
        newTileSize.h = Math.round(this.map.tileSize.h * Math.pow(2, newZoomlevel));

        // handles bug in ie: 
        // turn zoomOut tile scale off, if tileSize > 2^15 
        var ieBug;
        if (navigator.appName == "Microsoft Internet Explorer"){ 
            if ((newTileSize.w > Math.pow(2,15)) &&
                (newTileSize.h > Math.pow(2,15)) ){
                ieBug = true;
                this.map.baseLayer.zoomOutTile.imgDiv.style.display = "none";
            }
            else {
                ieBug = false;
            }
        }
        else {
            ieBug = false;
        }

        if (!ieBug && !this.firstCall) {
            //set new size (width and height)
            this.zoomOutTile.imgDiv.style.width = newTileSize.w + "px";          
            this.zoomOutTile.imgDiv.style.height = newTileSize.h + "px";

            //set new position (top and left)
            var bounds = this.zoomOutTile.bounds;
            var extent = this.map.getExtent();
            var maxExtent = this.map.getMaxExtent();

            var resolution = this.map.getMaxResolution();
            var centerLonLat = extent.getCenterLonLat();
             
            var offsetlon = bounds.left - centerLonLat.lon;
            var offsetlat = -bounds.top + centerLonLat.lat;
            
            var offsetx = offsetlon / newResolution
              - parseInt(this.map.layerContainerDiv.style.left);
            var offsety = offsetlat / newResolution
              - parseInt(this.map.layerContainerDiv.style.top);
            
            this.zoomOutTile.imgDiv.style.left = 
                Math.round(centerPx.x + offsetx) + "px";  
            this.zoomOutTile.imgDiv.style.top = 
                Math.round(centerPx.y + offsety) + "px";
        
            
            this.map.baseLayer.zoomOutTile.imgDiv.style.display = "block";
        }
        
        return true;
    },

    /** 
     * Clones layerContainer for "smooth tile update".
     * So, it gets no blank map while map is loading in new zoomlevel.
     *
     * Default function; will override in subclasses,
     * otherwise it returns false
     * 
     * @returns false if no layerContainer cloned
     * @type Boolean
     */
    cloneBaseLayerDiv:function() {
        return false;
    },

    /** 
     * The actual clone function:
     * Hold scaled map while tiles are loading successive.
     *  
     * special function; shares with some baselayers; calls from subclasses
     *
     */
    cloneBaseLayerDiv_share:function() {
        // function for map only; not for overviewmap!
        if (this.map.div.id == "map") {  
            
            // 1. clone baseLayerDiv with all childs (replace, if already exist)
            if (this.map.baseLayerDivClone){
                this.map.layerContainerDiv.removeChild(this.map.baseLayerDivClone);
                this.map.baseLayerDivClone = null;
            }
            this.map.baseLayerDivClone = this.map.baseLayer.div.cloneNode(true);
            this.map.baseLayerDivClone.id =
                this.map.baseLayer.div.id + "_clone";
                    
    
            // 2. append baseLayerDivClone to layerContainer
            // (now the cloned div is above the original; it hides the original)
            this.map.layerContainerDiv.appendChild(this.map.baseLayerDivClone);

            // 3. set zIndex 
            this.map.baseLayerDivClone.style.zIndex--; 
                
            // 4. set zoomOut tile (of original div) invisible 
            if (this.map.baseLayer.zoomOutTile)
                this.map.baseLayer.zoomOutTile.imgDiv.style.display = "none";

            // 5. adjust left and top position 
            // (if map is panning before, layerContainer is displaced)
            if (this.map.layerContainerDiv.style.left != "")
                this.map.baseLayerDivClone.style.left += 
                    parseInt(this.map.layerContainerDiv.style.left);
            if (this.map.layerContainerDiv.style.top != "")
                this.map.baseLayerDivClone.style.top += 
                    parseInt(this.map.layerContainerDiv.style.top);           
        }
    },


    /**
     * Sets visibility of map after zoomend/loadend:
     * - freezed/cloned layerContainer invisible
     * 
     * Calls automatically after fire loadend event. 
     */
    setLoadendVisibility: function() {
        // set flag that zoom animation finished
        this.map.zoomanimationActive = false;

        // function for map only; not for overviewmap!
        if (this.map.div.id == "map") {        
            // remove cloned baseLayerDiv
            if (this.map.baseLayerDivClone){
                this.map.layerContainerDiv.removeChild(this.map.baseLayerDivClone);
                this.map.baseLayerDivClone = null;
            }
        }
        
        // after map is loaded the first time the flag sets to false
        // (reason: zoomOutTile shouldn't visible by first zoom/scale)
        if (this.firstCall)
            this.firstCall = false;
    },


    /** @final @type String */
    CLASS_NAME: "OpenLayers.Layer"
};
